/*
 * Copyright (C) 2011 Atmark Techno, Inc. All Rights Reserved.
 */

/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/fsl_devices.h>
#include <linux/gpio.h>
#include <mach/hardware.h>
#include <mach/common.h>
#include "armadillo400_iomux.h"
#include "devices.h"
#include "board-armadillo400.h"

static struct pad_desc armadillo400_pwm1_pads[] = {
	MX25_PAD_PWM__PWM(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_UP),
};
static struct pad_desc armadillo400_pwm1_gpio_pads[] = {
	MX25_PAD_PWM__GPIO_1_26(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_UP),
};
static struct pad_desc __maybe_unused armadillo400_pwm2_pads[] = {
#if defined(CONFIG_ARMADILLO400_CON9_25_PWMO2)
	MX25_PAD_CSPI1_SS0__PWM2(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_UP),
#endif
};

static struct pad_desc __maybe_unused armadillo400_pwm3_pads[] = {
};

static struct pad_desc __maybe_unused armadillo400_pwm4_pads[] = {
#if defined(CONFIG_ARMADILLO400_CON14_3_PWMO4)
	MX25_PAD_GPIO_C__PWM4(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_UP),
#endif
};

static void __maybe_unused armadillo400_pwm3_enable_pad(void)
{
	mxc_iomux_v3a_setup_multiple_pads(armadillo400_pwm4_pads);
}

static void __maybe_unused armadillo400_pwm1_enable_pad(void)
{
	mxc_iomux_v3a_setup_multiple_pads(armadillo400_pwm2_pads);
}

static void armadillo400_pwm0_enable_pad(void)
{
	mxc_iomux_v3a_setup_multiple_pads(armadillo400_pwm1_pads);
}

static void armadillo400_pwm0_disable_pad(void)
{
	gpio_direction_output(PWM1_GPIO, 1);
	mxc_iomux_v3a_setup_multiple_pads(armadillo400_pwm1_gpio_pads);
}

static struct mxc_pwm_platform_data __maybe_unused armadillo400_pwm3_data = {
	.enable_pwm_pad = armadillo400_pwm3_enable_pad,
};

static struct mxc_pwm_platform_data __maybe_unused armadillo400_pwm1_data = {
	.enable_pwm_pad = armadillo400_pwm1_enable_pad,
};

static struct mxc_pwm_platform_data armadillo400_pwm0_data = {
	.pwmo_invert = 1,
	.enable_pwm_pad = armadillo400_pwm0_enable_pad,
	.disable_pwm_pad = armadillo400_pwm0_disable_pad,
};

void __init armadillo400_pwm_init(void)
{
	gpio_request(PWM1_GPIO, "pwm1_gpio");
	mxc_register_device(&mxc_pwm_device0, &armadillo400_pwm0_data);

#if defined(CONFIG_ARMADILLO400_CON9_25_PWMO2)
	mxc_register_device(&mxc_pwm_device1, &armadillo400_pwm1_data);
#endif

#if defined(CONFIG_ARMADILLO400_CON14_3_PWMO4)
	mxc_register_device(&mxc_pwm_device3, &armadillo400_pwm3_data);
#endif
}
